;;; WINDOWS-1250

;; #### this needs to move somewhere appropriate and be Custom-ized.
(defvar replacement-character ?\  ; yup, there's a space character
  "Define the character to be substituted for unknown character codes.
May be a character, or a list containing a charset or charset name,
followed by one or more octets determining a character in that set.

Intended for use as element of argument of `mapping-table-to-vector'
to generate CCL decode tables, usually inside a backquote construct in
the form `,replacement-character'.")

;; #### this needs to move somewhere appropriate
(defun mapping-table-to-vector (table)
  "Map TABLE to a 256-character vector.
Used to define CCL decode tables.

TABLE is a list.

Each element of TABLE may be a character, charset, natural number, or
a list of a charset followed by one or more natural numbers.  If a
character, it is inserted in the table at that position.  If a
charset, it must be a single-byte encoding, and it determines the
default charset to be used in constructing characters from all
following numerical codes until a new charset is set.  If a natural
number, it is converted to a character using the current default
charset, and inserted in the table at that position.  If a list,
`make-char' is called with the list members as arguments, leaving the
default charset unchanged.

The variable `replacement-character' contains a definition for the
\"undefined\" character.  See its docstring for usage.

If the element is not a charset, character, or natural number, an
error is signaled.

Multi-byte characters may be handled directly as characters, or using
the list format."
  (cond ((characterp replacement-character))
        ((and (consp replacement-character)
	      (condition-case nil
		  (apply #'make-char replacement-character)
		('error nil))))
	(t (error 'syntax-error
		  "`replacement-character' does not describe a character"
		  replacement-character)))
  (let ((cs 'ascii))
    (apply #'vector
	   (delq nil
		 (mapcar #'(lambda (code)
			     (let ((cs2 (if (symbolp code)
					    (find-charset code)
					  nil)))
			       (cond ((charsetp cs2) (setq cs cs2) nil)
				     ((characterp code) code)
				     ((consp code) (apply #'make-char code))
				     ((natnump code)
				      ;; #### Reset high bit kludge must die
				      ;; #### How about multibyte charsets?
				      (make-char cs (if (< code 128)
							code
						      (- code 128))))
				     (t (signal 'error
						`("Not a charset" ,code))))))
			 table)))))

(eval-and-compile

(defconst latin-windows-1250-decode-table
  ;; bind some convenient abbreviations for the table
  (let ((rc replacement-character)
	(ascii 'ascii)
	(latin-1 'latin-iso8859-1)
	(latin-2 'latin-iso8859-2))
    (mapping-table-to-vector `(
;; ISO8859-2  1250  UCS2  UISOname
,ascii
          0  ;  00  0000  NULL
          1  ;  01  0001  START OF HEADING
          2  ;  02  0002  START OF TEXT
          3  ;  03  0003  END OF TEXT
          4  ;  04  0004  END OF TRANSMISSION
          5  ;  05  0005  ENQUIRY
          6  ;  06  0006  ACKNOWLEDGE
          7  ;  07  0007  BELL
          8  ;  08  0008  BACKSPACE
          9  ;  09  0009  HORIZONTAL TABULATION
         10  ;  0A  000A  LINE FEED
         11  ;  0B  000B  VERTICAL TABULATION
         12  ;  0C  000C  FORM FEED
         13  ;  0D  000D  CARRIAGE RETURN
         14  ;  0E  000E  SHIFT OUT
         15  ;  0F  000F  SHIFT IN
         16  ;  10  0010  DATA LINK ESCAPE
         17  ;  11  0011  DEVICE CONTROL ONE
         18  ;  12  0012  DEVICE CONTROL TWO
         19  ;  13  0013  DEVICE CONTROL THREE
         20  ;  14  0014  DEVICE CONTROL FOUR
         21  ;  15  0015  NEGATIVE ACKNOWLEDGE
         22  ;  16  0016  SYNCHRONOUS IDLE
         23  ;  17  0017  END OF TRANSMISSION BLOCK
         24  ;  18  0018  CANCEL
         25  ;  19  0019  END OF MEDIUM
         26  ;  1A  001A  SUBSTITUTE
         27  ;  1B  001B  ESCAPE
         28  ;  1C  001C  FILE SEPARATOR
         29  ;  1D  001D  GROUP SEPARATOR
         30  ;  1E  001E  RECORD SEPARATOR
         31  ;  1F  001F  UNIT SEPARATOR
         32  ;  20  0020  SPACE
         33  ;  21  0021  EXCLAMATION MARK
         34  ;  22  0022  QUOTATION MARK
         35  ;  23  0023  NUMBER SIGN
         36  ;  24  0024  DOLLAR SIGN
         37  ;  25  0025  PERCENT SIGN
         38  ;  26  0026  AMPERSAND
         39  ;  27  0027  APOSTROPHE
         40  ;  28  0028  LEFT PARENTHESIS
         41  ;  29  0029  RIGHT PARENTHESIS
         42  ;  2A  002A  ASTERISK
         43  ;  2B  002B  PLUS SIGN
         44  ;  2C  002C  COMMA
         45  ;  2D  002D  HYPHEN-MINUS
         46  ;  2E  002E  FULL STOP
         47  ;  2F  002F  SOLIDUS
         48  ;  30  0030  DIGIT ZERO
         49  ;  31  0031  DIGIT ONE
         50  ;  32  0032  DIGIT TWO
         51  ;  33  0033  DIGIT THREE
         52  ;  34  0034  DIGIT FOUR
         53  ;  35  0035  DIGIT FIVE
         54  ;  36  0036  DIGIT SIX
         55  ;  37  0037  DIGIT SEVEN
         56  ;  38  0038  DIGIT EIGHT
         57  ;  39  0039  DIGIT NINE
         58  ;  3A  003A  COLON
         59  ;  3B  003B  SEMICOLON
         60  ;  3C  003C  LESS-THAN SIGN
         61  ;  3D  003D  EQUALS SIGN
         62  ;  3E  003E  GREATER-THAN SIGN
         63  ;  3F  003F  QUESTION MARK
         64  ;  40  0040  COMMERCIAL AT
         65  ;  41  0041  LATIN CAPITAL LETTER A
         66  ;  42  0042  LATIN CAPITAL LETTER B
         67  ;  43  0043  LATIN CAPITAL LETTER C
         68  ;  44  0044  LATIN CAPITAL LETTER D
         69  ;  45  0045  LATIN CAPITAL LETTER E
         70  ;  46  0046  LATIN CAPITAL LETTER F
         71  ;  47  0047  LATIN CAPITAL LETTER G
         72  ;  48  0048  LATIN CAPITAL LETTER H
         73  ;  49  0049  LATIN CAPITAL LETTER I
         74  ;  4A  004A  LATIN CAPITAL LETTER J
         75  ;  4B  004B  LATIN CAPITAL LETTER K
         76  ;  4C  004C  LATIN CAPITAL LETTER L
         77  ;  4D  004D  LATIN CAPITAL LETTER M
         78  ;  4E  004E  LATIN CAPITAL LETTER N
         79  ;  4F  004F  LATIN CAPITAL LETTER O
         80  ;  50  0050  LATIN CAPITAL LETTER P
         81  ;  51  0051  LATIN CAPITAL LETTER Q
         82  ;  52  0052  LATIN CAPITAL LETTER R
         83  ;  53  0053  LATIN CAPITAL LETTER S
         84  ;  54  0054  LATIN CAPITAL LETTER T
         85  ;  55  0055  LATIN CAPITAL LETTER U
         86  ;  56  0056  LATIN CAPITAL LETTER V
         87  ;  57  0057  LATIN CAPITAL LETTER W
         88  ;  58  0058  LATIN CAPITAL LETTER X
         89  ;  59  0059  LATIN CAPITAL LETTER Y
         90  ;  5A  005A  LATIN CAPITAL LETTER Z
         91  ;  5B  005B  LEFT SQUARE BRACKET
         92  ;  5C  005C  REVERSE SOLIDUS
         93  ;  5D  005D  RIGHT SQUARE BRACKET
         94  ;  5E  005E  CIRCUMFLEX ACCENT
         95  ;  5F  005F  LOW LINE
         96  ;  60  0060  GRAVE ACCENT
         97  ;  61  0061  LATIN SMALL LETTER A
         98  ;  62  0062  LATIN SMALL LETTER B
         99  ;  63  0063  LATIN SMALL LETTER C
        100  ;  64  0064  LATIN SMALL LETTER D
        101  ;  65  0065  LATIN SMALL LETTER E
        102  ;  66  0066  LATIN SMALL LETTER F
        103  ;  67  0067  LATIN SMALL LETTER G
        104  ;  68  0068  LATIN SMALL LETTER H
        105  ;  69  0069  LATIN SMALL LETTER I
        106  ;  6A  006A  LATIN SMALL LETTER J
        107  ;  6B  006B  LATIN SMALL LETTER K
        108  ;  6C  006C  LATIN SMALL LETTER L
        109  ;  6D  006D  LATIN SMALL LETTER M
        110  ;  6E  006E  LATIN SMALL LETTER N
        111  ;  6F  006F  LATIN SMALL LETTER O
        112  ;  70  0070  LATIN SMALL LETTER P
        113  ;  71  0071  LATIN SMALL LETTER Q
        114  ;  72  0072  LATIN SMALL LETTER R
        115  ;  73  0073  LATIN SMALL LETTER S
        116  ;  74  0074  LATIN SMALL LETTER T
        117  ;  75  0075  LATIN SMALL LETTER U
        118  ;  76  0076  LATIN SMALL LETTER V
        119  ;  77  0077  LATIN SMALL LETTER W
        120  ;  78  0078  LATIN SMALL LETTER X
        121  ;  79  0079  LATIN SMALL LETTER Y
        122  ;  7A  007A  LATIN SMALL LETTER Z
        123  ;  7B  007B  LEFT CURLY BRACKET
        124  ;  7C  007C  VERTICAL LINE
        125  ;  7D  007D  RIGHT CURLY BRACKET
        126  ;  7E  007E  TILDE
        127  ;  7F  007F  DELETE
,latin-2
        ,rc  ;  80  20AC  EURO SIGN
        ,rc  ;  81  ;     UNDEFINED
        ,rc  ;  82  201A  SINGLE LOW-9 QUOTATION MARK
        ,rc  ;  83  ;     UNDEFINED
        ,rc  ;  84  201E  DOUBLE LOW-9 QUOTATION MARK
        ,rc  ;  85  2026  HORIZONTAL ELLIPSIS
        ,rc  ;  86  2020  DAGGER
        ,rc  ;  87  2021  DOUBLE DAGGER
        ,rc  ;  88  ;     UNDEFINED
        ,rc  ;  89  2030  PER MILLE SIGN
        169  ;  8A  0160  LATIN CAPITAL LETTER S WITH CARON
        ,rc  ;  8B  2039  SINGLE LEFT-POINTING ANGLE QUOTATION MARK
        166  ;  8C  015A  LATIN CAPITAL LETTER S WITH ACUTE
        171  ;  8D  0164  LATIN CAPITAL LETTER T WITH CARON
        174  ;  8E  017D  LATIN CAPITAL LETTER Z WITH CARON
        172  ;  8F  0179  LATIN CAPITAL LETTER Z WITH ACUTE
        ,rc  ;  90  ;     UNDEFINED
        ,rc  ;  91  2018  LEFT SINGLE QUOTATION MARK
        ,rc  ;  92  2019  RIGHT SINGLE QUOTATION MARK
        ,rc  ;  93  201C  LEFT DOUBLE QUOTATION MARK
        ,rc  ;  94  201D  RIGHT DOUBLE QUOTATION MARK
        ,rc  ;  95  2022  BULLET
        ,rc  ;  96  2013  EN DASH
        ,rc  ;  97  2014  EM DASH
        ,rc  ;  98  ;     UNDEFINED
        ,rc  ;  99  2122  TRADE MARK SIGN
        185  ;  9A  0161  LATIN SMALL LETTER S WITH CARON
        ,rc  ;  9B  203A  SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
        182  ;  9C  015B  LATIN SMALL LETTER S WITH ACUTE
        187  ;  9D  0165  LATIN SMALL LETTER T WITH CARON
        190  ;  9E  017E  LATIN SMALL LETTER Z WITH CARON
        188  ;  9F  017A  LATIN SMALL LETTER Z WITH ACUTE
         32  ;  A0  00A0  NO-BREAK SPACE
         55  ;  A1  02C7  CARON
         34  ;  A2  02D8  BREVE
         35  ;  A3  0141  LATIN CAPITAL LETTER L WITH STROKE
         36  ;  A4  00A4  CURRENCY SIGN
         33  ;  A5  0104  LATIN CAPITAL LETTER A WITH OGONEK
(,latin-1 38) ;  A6  00A6  BROKEN BAR
         39  ;  A7  00A7  SECTION SIGN
         40  ;  A8  00A8  DIAERESIS
(,latin-1 41) ;  A9  00A9  COPYRIGHT SIGN
         42  ;  AA  015E  LATIN CAPITAL LETTER S WITH CEDILLA
(,latin-1 43) ;  AB  00AB  LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
(,latin-1 44) ;  AC  00AC  NOT SIGN
         45  ;  AD  00AD  SOFT HYPHEN
(,latin-1 46) ;  AE  00AE  REGISTERED SIGN
         47  ;  AF  017B  LATIN CAPITAL LETTER Z WITH DOT ABOVE
         48  ;  B0  00B0  DEGREE SIGN
(,latin-1 49) ;  B1  00B1  PLUS-MINUS SIGN
         50  ;  B2  02DB  OGONEK
         51  ;  B3  0142  LATIN SMALL LETTER L WITH STROKE
         52  ;  B4  00B4  ACUTE ACCENT
(,latin-1 53) ;  B5  00B5  MICRO SIGN
(,latin-1 54) ;  B6  00B6  PILCROW SIGN
(,latin-1 55) ;  B7  00B7  MIDDLE DOT
         56  ;  B8  00B8  CEDILLA
         49  ;  B9  0105  LATIN SMALL LETTER A WITH OGONEK
         58  ;  BA  015F  LATIN SMALL LETTER S WITH CEDILLA
(,latin-1 59) ;  BB  00BB  RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
         37  ;  BC  013D  LATIN CAPITAL LETTER L WITH CARON
         61  ;  BD  02DD  DOUBLE ACUTE ACCENT
         53  ;  BE  013E  LATIN SMALL LETTER L WITH CARON
         63  ;  BF  017C  LATIN SMALL LETTER Z WITH DOT ABOVE
         64  ;  C0  0154  LATIN CAPITAL LETTER R WITH ACUTE
         65  ;  C1  00C1  LATIN CAPITAL LETTER A WITH ACUTE
         66  ;  C2  00C2  LATIN CAPITAL LETTER A WITH CIRCUMFLEX
         67  ;  C3  0102  LATIN CAPITAL LETTER A WITH BREVE
         68  ;  C4  00C4  LATIN CAPITAL LETTER A WITH DIAERESIS
         69  ;  C5  0139  LATIN CAPITAL LETTER L WITH ACUTE
         70  ;  C6  0106  LATIN CAPITAL LETTER C WITH ACUTE
         71  ;  C7  00C7  LATIN CAPITAL LETTER C WITH CEDILLA
         72  ;  C8  010C  LATIN CAPITAL LETTER C WITH CARON
         73  ;  C9  00C9  LATIN CAPITAL LETTER E WITH ACUTE
         74  ;  CA  0118  LATIN CAPITAL LETTER E WITH OGONEK
         75  ;  CB  00CB  LATIN CAPITAL LETTER E WITH DIAERESIS
         76  ;  CC  011A  LATIN CAPITAL LETTER E WITH CARON
         77  ;  CD  00CD  LATIN CAPITAL LETTER I WITH ACUTE
         78  ;  CE  00CE  LATIN CAPITAL LETTER I WITH CIRCUMFLEX
         79  ;  CF  010E  LATIN CAPITAL LETTER D WITH CARON
         80  ;  D0  0110  LATIN CAPITAL LETTER D WITH STROKE
         81  ;  D1  0143  LATIN CAPITAL LETTER N WITH ACUTE
         82  ;  D2  0147  LATIN CAPITAL LETTER N WITH CARON
         83  ;  D3  00D3  LATIN CAPITAL LETTER O WITH ACUTE
         84  ;  D4  00D4  LATIN CAPITAL LETTER O WITH CIRCUMFLEX
         85  ;  D5  0150  LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
         86  ;  D6  00D6  LATIN CAPITAL LETTER O WITH DIAERESIS
         87  ;  D7  00D7  MULTIPLICATION SIGN
         88  ;  D8  0158  LATIN CAPITAL LETTER R WITH CARON
         89  ;  D9  016E  LATIN CAPITAL LETTER U WITH RING ABOVE
         90  ;  DA  00DA  LATIN CAPITAL LETTER U WITH ACUTE
         91  ;  DB  0170  LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
         92  ;  DC  00DC  LATIN CAPITAL LETTER U WITH DIAERESIS
         93  ;  DD  00DD  LATIN CAPITAL LETTER Y WITH ACUTE
         94  ;  DE  0162  LATIN CAPITAL LETTER T WITH CEDILLA
         95  ;  DF  00DF  LATIN SMALL LETTER SHARP S
         96  ;  E0  0155  LATIN SMALL LETTER R WITH ACUTE
         97  ;  E1  00E1  LATIN SMALL LETTER A WITH ACUTE
         98  ;  E2  00E2  LATIN SMALL LETTER A WITH CIRCUMFLEX
         99  ;  E3  0103  LATIN SMALL LETTER A WITH BREVE
        100  ;  E4  00E4  LATIN SMALL LETTER A WITH DIAERESIS
        101  ;  E5  013A  LATIN SMALL LETTER L WITH ACUTE
        102  ;  E6  0107  LATIN SMALL LETTER C WITH ACUTE
        103  ;  E7  00E7  LATIN SMALL LETTER C WITH CEDILLA
        104  ;  E8  010D  LATIN SMALL LETTER C WITH CARON
        105  ;  E9  00E9  LATIN SMALL LETTER E WITH ACUTE
        106  ;  EA  0119  LATIN SMALL LETTER E WITH OGONEK
        107  ;  EB  00EB  LATIN SMALL LETTER E WITH DIAERESIS
        108  ;  EC  011B  LATIN SMALL LETTER E WITH CARON
        109  ;  ED  00ED  LATIN SMALL LETTER I WITH ACUTE
        110  ;  EE  00EE  LATIN SMALL LETTER I WITH CIRCUMFLEX
        111  ;  EF  010F  LATIN SMALL LETTER D WITH CARON
        112  ;  F0  0111  LATIN SMALL LETTER D WITH STROKE
        113  ;  F1  0144  LATIN SMALL LETTER N WITH ACUTE
        114  ;  F2  0148  LATIN SMALL LETTER N WITH CARON
        115  ;  F3  00F3  LATIN SMALL LETTER O WITH ACUTE
        116  ;  F4  00F4  LATIN SMALL LETTER O WITH CIRCUMFLEX
        117  ;  F5  0151  LATIN SMALL LETTER O WITH DOUBLE ACUTE
        118  ;  F6  00F6  LATIN SMALL LETTER O WITH DIAERESIS
        119  ;  F7  00F7  DIVISION SIGN
        120  ;  F8  0159  LATIN SMALL LETTER R WITH CARON
        121  ;  F9  016F  LATIN SMALL LETTER U WITH RING ABOVE
        122  ;  FA  00FA  LATIN SMALL LETTER U WITH ACUTE
        123  ;  FB  0171  LATIN SMALL LETTER U WITH DOUBLE ACUTE
        124  ;  FC  00FC  LATIN SMALL LETTER U WITH DIAERESIS
        125  ;  FD  00FD  LATIN SMALL LETTER Y WITH ACUTE
        126  ;  FE  0163  LATIN SMALL LETTER T WITH CEDILLA
        127  ;  FF  02D9  DOT ABOVE
   )))
  "Latin Windows-1250 decoding table.")

(defconst latin-windows-1250-encode-table
  (let ((table (make-vector 256 32))
	(i 0))
    (while (< i 256)
      (let* ((ch (aref latin-windows-1250-decode-table i))
	     (split (split-char ch)))
	(cond ((eq (car split) 'latin-iso8859-2)
	       (aset table (logior (nth 1 split) 128) i)
	       )
	      ((eq ch 32))
	      ((eq (car split) 'ascii)
	       (aset table ch i)
	       )))
      (setq i (1+ i)))
    table)
  "Latin Windows-1250 encoding table.")

)

(define-ccl-program ccl-decode-windows1250
  `(3
    ((read r0)
     (loop
      (write-read-repeat r0 ,latin-windows-1250-decode-table))))
  "CCL program to decode Windows-1250.")

(define-ccl-program ccl-encode-windows1250
  `(1
    ((read r0)
     (loop
      (if (r0 != ,(charset-id 'latin-iso8859-2))
	  (write-read-repeat r0)
	((read r0)
	 (write-read-repeat r0 ,latin-windows-1250-encode-table))))))
  "CCL program to encode Windows-1250.")

(make-coding-system
 'windows-1250 'ccl
 "Coding-system used for Windows-1250."
 `(decode ,ccl-decode-windows1250
   encode nil ; ,ccl-encode-windows1250
   mnemonic "W1250"))

;; `iso-8-1' is not correct, but XEmacs doesn't have a `ccl' category
(coding-system-put 'windows-1250 'category 'iso-8-1)

;; This is really bogus and you shouldn't need it but just in case.
;(set-language-info-alist
; "Latin-2-Win" '((charset latin-iso8859-2)
;		   (coding-system windows-1250)
;		   (coding-priority windows-1250)
;		   (input-method . "latin-2-prefix")
;		   ;(features latin-2-util)
;		   (tutorial . "TUTORIAL.hr")
;		   (sample-text . "Croatian")
;		   (documentation . "Support for Latin Windows-1250."))
; '("Latin-2-Nonsense"))

